import itertools
import itertools
import re
import sys
from functools import reduce
from math import ceil, floor, log

from checkov.terraform.parser_functions import tonumber, FUNCTION_FAILED, create_map, tobool, tomap, tostring

"""
This file contains a custom implementation of the builtin `eval` function.
`eval` is not a safe function, because it can execute *every* command, 
so this file overrides `eval` and allows only the functions in SAFE_EVAL_DICT.

The functions are an implementation of Terraform's built-in functions
https://www.terraform.io/docs/configuration/functions.html

Not all of the functions are implemented yet. If a function doesn't exist, the original value is returned.
"""


def _find_regex_groups(pattern, input_str):
    match = re.match(pattern, input_str)
    if match:
        if match.groupdict():
            # try to find named capturing groups
            return match.groupdict()
        if list(match.groups()):
            # try to find unnamed capturing groups
            return list(match.groups())
    return None


def regex(pattern, input_str):
    try:
        groups = _find_regex_groups(pattern, input_str)
        if groups is not None:
            return groups

        results = re.findall(pattern, input_str)
        # return first match
        if len(results) > 0:
            return results[0]
        return ''
    except TypeError:
        return f'regex({pattern}, {input_str})'


def regexall(pattern, input_str):
    try:
        groups = _find_regex_groups(pattern, input_str)
        if groups is not None:
            return groups

        results = re.findall(pattern, input_str)
        return results
    except TypeError:
        return f'regexall({pattern}, {input_str})'


def trim(input_str, chars_to_remove):
    for c in chars_to_remove:
        input_str = input_str.replace(c, '')
    return input_str


def coalesce(*arg):
    return reduce(lambda x, y: x if x not in [None, ""] else y, arg)


def coalesce_list(*arg):
    return reduce(lambda x, y: x if x not in [None, []] else y, arg)


def flatten(lst):
    res = [item for sublist in lst for item in sublist]
    if any(type(elem) is list for elem in res):
        return flatten(res)
    else:
        return res


def matchkeys(values_list, keys_list, search_set):
    matching = set()
    for search in search_set:
        indices = [i for i, x in enumerate(keys_list) if x == search]
        for i in indices:
            matching.add(values_list[i])

    return list(matching)


def reverse(lst):
    lst.reverse()
    return lst


def sort(lst):
    lst.sort()
    return lst


def merge(*args):
    res = {}
    for d in args:
        res = {**res, **d}
    return res


def wrap_func(f, *args):
    res = f(*args)
    if res == FUNCTION_FAILED:
        raise ValueError
    return res


SAFE_EVAL_FUNCTIONS = []
SAFE_EVAL_DICT = dict([(k, locals().get(k, None)) for k in SAFE_EVAL_FUNCTIONS])

# math functions
SAFE_EVAL_DICT['abs'] = abs
SAFE_EVAL_DICT['ceil'] = ceil
SAFE_EVAL_DICT['floor'] = floor
SAFE_EVAL_DICT['log'] = log
SAFE_EVAL_DICT['max'] = max
SAFE_EVAL_DICT['min'] = min
SAFE_EVAL_DICT['parsint'] = int
SAFE_EVAL_DICT['pow'] = pow
SAFE_EVAL_DICT['signum'] = lambda x: -1 if x < 0 else 0 if x == 0 else 1

# string functions
SAFE_EVAL_DICT['chomp'] = lambda x: x.rstrip()
SAFE_EVAL_DICT['format'] = lambda text_to_format, *args: (text_to_format % args)
SAFE_EVAL_DICT['formatlist'] = lambda text_to_format, args_list: [(text_to_format % args) for args in args_list]
SAFE_EVAL_DICT['indent'] = lambda num_of_space, input_str: input_str
SAFE_EVAL_DICT['join'] = lambda separator, lst: separator.join(lst)
SAFE_EVAL_DICT['lower'] = lambda input_str: input_str.lower()
SAFE_EVAL_DICT['regex'] = regex
SAFE_EVAL_DICT['regexall'] = regexall
SAFE_EVAL_DICT['replace'] = lambda string, substring, replacement: string.replace(substring, replacement)
SAFE_EVAL_DICT['split'] = lambda separator, input_str: input_str.split(separator)
SAFE_EVAL_DICT['strrev'] = lambda input_str: input_str[::-1]
SAFE_EVAL_DICT['substr'] = lambda input_str, offset, length: input_str[offset:offset + length]
SAFE_EVAL_DICT['title'] = lambda input_str: input_str.title()
SAFE_EVAL_DICT['trim'] = trim
SAFE_EVAL_DICT['trimprefix'] = lambda input_str, prefix: input_str.lstrip(prefix)
SAFE_EVAL_DICT['trimsuffix'] = lambda input_str, prefix: input_str.rstrip(prefix)
SAFE_EVAL_DICT['trimspace'] = lambda input_str: input_str.strip()
SAFE_EVAL_DICT['upper'] = lambda input_str: input_str.upper()

# collections
SAFE_EVAL_DICT['chunklist'] = lambda lst, chunk_size: [lst[i:i + chunk_size] for i in range(0, len(lst), chunk_size)]
SAFE_EVAL_DICT['coalesce'] = coalesce
SAFE_EVAL_DICT['coalescelist'] = coalesce_list
SAFE_EVAL_DICT['compact'] = lambda lst: list(filter(lambda l: l != "", lst))
SAFE_EVAL_DICT['concat'] = lambda *lists: list(itertools.chain(*lists))
SAFE_EVAL_DICT['contains'] = lambda lst, value: value in lst
SAFE_EVAL_DICT['distinct'] = lambda lst: list(dict.fromkeys(lst))
SAFE_EVAL_DICT['element'] = lambda lst, index: lst[index]
SAFE_EVAL_DICT['flatten'] = flatten
SAFE_EVAL_DICT['index'] = lambda lst, value: lst.index(value)
SAFE_EVAL_DICT['keys'] = lambda map_input: list(map_input.keys())
SAFE_EVAL_DICT['length'] = len
SAFE_EVAL_DICT['list'] = lambda *args: list(args)
SAFE_EVAL_DICT['lookup'] = lambda map_input, key, default: map_input.get(key, default)
SAFE_EVAL_DICT['map'] = lambda *args: wrap_func(create_map, list(args))
SAFE_EVAL_DICT['matchkeys'] = matchkeys
SAFE_EVAL_DICT['merge'] = merge
# SAFE_EVAL_DICT['range']
SAFE_EVAL_DICT['reverse'] = reverse
SAFE_EVAL_DICT['sort'] = sort


# type conversion
SAFE_EVAL_DICT['tobool'] = lambda arg: wrap_func(tobool, arg)
SAFE_EVAL_DICT['tolist'] = lambda *args: list(*args)
SAFE_EVAL_DICT['tomap'] = lambda arg: wrap_func(tomap, str(arg))
SAFE_EVAL_DICT['tonumber'] = lambda arg: arg if type(arg) in [int, float] else wrap_func(tonumber, arg)
SAFE_EVAL_DICT['toset'] = lambda origin: set(origin)
SAFE_EVAL_DICT['tostring'] = lambda arg: arg if isinstance(arg, str) else wrap_func(tostring, str(arg))

# encoding
SAFE_EVAL_DICT['jsonencode'] = lambda arg: arg


def get_allowed_functions():
    return list(SAFE_EVAL_DICT.keys())


class PythonCodeError(ValueError):
    # Create a custom error class for usage in python builtins
    pass


def evaluate(input_str):
    if "__" in input_str:
        err_msg = f"got a substring with double underscore, which is not allowed. origin string: {input_str}"
        raise PythonCodeError(err_msg)
    return eval(input_str, {"__builtins__": None}, SAFE_EVAL_DICT)  # nosec




from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck
from checkov.common.models.enums import CheckCategories, CheckResult


class GoogleCloudDNSSECEnabled(BaseResourceValueCheck):
    """
    Looks for DNSSEC state at dns_managed_zone:
    https://www.terraform.io/docs/providers/google/r/dns_managed_zone.html#state
    """

    def __init__(self):
        name = "Ensure that DNSSEC is enabled for Cloud DNS"
        id = "CKV_GCP_16"
        supported_resources = ["google_dns_managed_zone"]
        categories = [CheckCategories.ENCRYPTION]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):

        if 'visibility' in conf:
            if conf['visibility'][0] == 'private':
                return CheckResult.UNKNOWN  # check is irrelevant (cannot set DNSSEC to anything else)

        # default visibility is public; just use base class implementation
        return super().scan_resource_conf(conf)

    def get_inspected_key(self):
        return "dnssec_config/[0]/state"

    def get_expected_value(self):
        return "on"

    def get_expected_values(self):
        return [self.get_expected_value(), "transfer"]


check = GoogleCloudDNSSECEnabled()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.dockerfile.base_dockerfile_check import BaseDockerfileCheck


class RootUser(BaseDockerfileCheck):
    def __init__(self):
        name = "Ensure the last USER is not root"
        id = "CKV_DOCKER_8"
        supported_instructions = ["USER"]
        categories = [CheckCategories.APPLICATION_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_instructions=supported_instructions)

    def scan_entity_conf(self, conf):
        last_user = conf[-1]
        if last_user["value"] == "root":
            return CheckResult.FAILED, last_user

        return CheckResult.PASSED, last_user


check = RootUser()



from checkov.common.graph.checks_infra.enums import Operators
from .ending_with_attribute_solver import EndingWithAttributeSolver


class NotEndingWithAttributeSolver(EndingWithAttributeSolver):
    operator = Operators.NOT_ENDING_WITH

    def __init__(self, resource_types, attribute, value):
        super().__init__(resource_types=resource_types,
                         attribute=attribute, value=value)

    def _get_operation(self, vertex, attribute):
        return not super(NotEndingWithAttributeSolver, self)._get_operation(vertex, attribute)



from typing import Dict, Any, Tuple, List

from checkov.common.checks.base_check_registry import BaseCheckRegistry


class Registry(BaseCheckRegistry):
    def __init__(self) -> None:
        super().__init__()

    def extract_entity_details(
        self, entity: Dict[str, Dict[str, Dict[str, List[Dict[str, Any]]]]]
    ) -> Tuple[str, str, Dict[str, List[Dict[str, Any]]]]:
        data_type, data_object = next(iter(entity.items()))
        data_name, data_configuration = next(iter(data_object.items()))
        return data_type, data_name, data_configuration



from dataclasses import dataclass

from checkov.common.graph.graph_builder import CustomAttributes as CommonCustomAttributes, props


@dataclass
class CustomAttributes(CommonCustomAttributes):
    ENCRYPTION = "encryption_"
    ENCRYPTION_DETAILS = "encryption_details_"


reserved_attribute_names = props(CustomAttributes)



from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.models.enums import CheckCategories, CheckResult


class KubernetesSecretsEncryptedUsingCMK(BaseResourceCheck):
    def __init__(self):
        name = "Ensure Kubernetes Secrets are encrypted using Customer Master Keys (CMKs) managed in AWS KMS"
        id = "CKV_AWS_151"
        supported_resources = ["aws_eks_cluster"]
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if "encryption_config" not in conf.keys():
            return CheckResult.FAILED
        else:
            encryption_config = conf.get("encryption_config")[0]
            if "resources" in encryption_config and "provider" in encryption_config:
                resources = encryption_config.get("resources")[0]
                provider = encryption_config.get("provider")[0]
                if isinstance(resources, list):
                    if "secrets" in resources and "key_arn" in provider:
                        return CheckResult.PASSED
                    return CheckResult.UNKNOWN

        return CheckResult.FAILED


check = KubernetesSecretsEncryptedUsingCMK()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.base_spec_check import BaseK8Check


class KubernetesDashboard(BaseK8Check):

    def __init__(self):
        name = "Ensure the Kubernetes dashboard is not deployed"
        id = "CKV_K8S_33"
        # Location: container .image
        supported_kind = ['containers', 'initContainers']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_kind)

    def get_resource_id(self, conf):
        return f'{conf["parent"]} - {conf["name"]}'

    def scan_spec_conf(self, conf):
        if "image" in conf:
            conf_image = conf["image"]
            if not isinstance(conf_image, str):
                return CheckResult.FAILED
            if ("kubernetes-dashboard" in conf_image or "kubernetesui" in conf_image):
                return CheckResult.FAILED
        else:
            return CheckResult.FAILED
        if "parent_metadata" in conf:
            if "labels" in conf["parent_metadata"]:
                if "app" in conf["parent_metadata"]["labels"]:
                    if conf["parent_metadata"]["labels"]["app"] == "kubernetes-dashboard":
                        return CheckResult.FAILED
                elif "k8s-app" in conf["parent_metadata"]["labels"]:
                    if conf["parent_metadata"]["labels"]["k8s-app"] == "kubernetes-dashboard":
                        return CheckResult.FAILED
        return CheckResult.PASSED


check = KubernetesDashboard()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck


class AutomationEncrypted(BaseResourceValueCheck):
    def __init__(self):
        name = "Ensure that Automation account variables are encrypted"
        id = "CKV_AZURE_73"
        supported_resources = ['azurerm_automation_variable_bool','azurerm_automation_variable_string', 'azurerm_automation_variable_int', 'azurerm_automation_variable_datetime']
        categories = [CheckCategories.ENCRYPTION]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self):
        return 'encrypted'


check = AutomationEncrypted()



from checkov.cloudformation.checks.resource.base_resource_value_check import BaseResourceValueCheck
from checkov.common.models.enums import CheckCategories


class ELBAccessLogs(BaseResourceValueCheck):
    def __init__(self):
        name = "Ensure the ELB has access logging enabled"
        id = "CKV_AWS_92"
        supported_resources = ['AWS::ElasticLoadBalancing::LoadBalancer']
        categories = [CheckCategories.LOGGING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self):
        return 'Properties/AccessLoggingPolicy/Enabled'


check = ELBAccessLogs()



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.arm.base_resource_check import BaseResourceCheck

# https://docs.microsoft.com/en-us/azure/templates/microsoft.keyvault/2016-10-01/vaults

class KeyVaultRecoveryEnabled(BaseResourceCheck):
    def __init__(self):
        name = "Ensure the key vault is recoverable"
        id = "CKV_AZURE_42"
        supported_resources = ['Microsoft.KeyVault/vaults']
        categories = [CheckCategories.BACKUP_AND_RECOVERY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        # NOTE: enablePurgeProtection not supported in API version 2015-06-01
        if "properties" in conf:
            if "enablePurgeProtection" in conf["properties"] and "enableSoftDelete" in conf["properties"]:
                if str(conf["properties"]["enablePurgeProtection"]).lower() == "true" and \
                        str(conf["properties"]["enableSoftDelete"]).lower() == "true":
                    return CheckResult.PASSED
        return CheckResult.FAILED

check = KeyVaultRecoveryEnabled()



from collections import OrderedDict

import re
from dockerfile_parse import DockerfileParser
from dockerfile_parse.constants import COMMENT_INSTRUCTION

# class CheckovDockerFileParser(DockerfileParser)
from checkov.common.comment.enum import COMMENT_REGEX


def parse(filename):
    dfp = DockerfileParser(path=filename)
    return dfp_group_by_instructions(dfp)


def dfp_group_by_instructions(dfp):
    result = OrderedDict()
    for instruction in dfp.structure:
        instruction_literal = instruction["instruction"]
        if instruction_literal not in result:
            result[instruction_literal] = []
        result[instruction_literal].append(instruction)
    return result, dfp.lines


def collect_skipped_checks(parse_result):
    skipped_checks = []
    if COMMENT_INSTRUCTION in parse_result:
        for comment in parse_result[COMMENT_INSTRUCTION]:
            skip_search = re.search(COMMENT_REGEX, comment["value"])
            if skip_search:
                skipped_checks.append(
                    {
                        'id': skip_search.group(2),
                        'suppress_comment': skip_search.group(3)[1:] if skip_search.group(
                            3) else "No comment provided"
                    }
                )
    return skipped_checks




from checkov.common.graph.checks_infra.enums import Operators
from checkov.terraform.checks_infra.solvers.connections_solvers.connection_exists_solver import ConnectionExistsSolver


class ConnectionNotExistsSolver(ConnectionExistsSolver):
    operator = Operators.NOT_EXISTS

    def __init__(self, resource_types, connected_resources_types, vertices_under_resource_types=None, vertices_under_connected_resources_types=None):
        super().__init__(resource_types, connected_resources_types, vertices_under_resource_types, vertices_under_connected_resources_types)

    def get_operation(self, graph_connector):
        passed, failed = super(ConnectionNotExistsSolver, self).get_operation(graph_connector)
        return failed, passed
